// Main file for the DDR SDRAM controller
//
// Memory timing:
// tREFI = 7.8uS	(Max Periodic Refresh Interval)
// With frequency 100MHz (period 10ns) it will be 30Ch counts
// Init of the memory can be started after 200us elapsed (4E20h counts)
// DDR SDRAM must be configured for sequential burst type
// As memory has 4 banks x 4M of 16bit words (8Mbytes each bank)
// address from VJTAG decoded as following:
// 0000000 - 07FFFFF - bank1
// 0800000 - 0FFFFFF - bank2
// 1000000 - 17FFFFF - bank3
// 1800000 - 1FFFFFF - bank4
// Raw address specified by a0:a12 during ACT command
// Column address selected by a0:a8 during READ/WRITE command
// a10 specifies AP (auto precharge) mode during READ/WRITE command
// a10 = 1 - AP enabled
// Address translation will be:
// {ba[1],ba[0],r[12]-r[0],c[8]-c[0],i}
// ba - bank
// r - raw 
// c - column
// i - ignored
// Pins assignment in ddr_cmd[3:0] register
// ddr_cmd[3] - cs
// ddr_cmd[2] - ras
// ddr_cmd[1] - cas
// ddr_cmd[0] - we
//
module ddr_cntr (
ddr_a, ddr_cmd, ddr_ba, ddr_clk, 
ddr_dq, ddr_dm, ddr_dqs,ddr_data_out, ddr_cke, ddr_test_out,
ddr_pll_in, ddr_pll_90, ddr_cmd_in, ddr_ba_in, ddr_r_in,
ddr_c_in, ddr_data_in, ddr_test_shift, ddr_test_clk
);
 

  `include "parameters.v"

  parameter READ_LT = 3'b101;   // Specifies duration of read cycle
  
   // MRS parameter written to the register {a12..a0} ba[1:0] = 2'b00;
   // Bits 12-9 = 0 (not used)
   // Bit 8 (DLL  Reset): 0 - No, 1 - Yes
   // Bit 7 = 0 (not used)
   // Bits 6-4 (LTMODE - CAS latency):  010 - 2
   //								 	011 - 3
   //								 	101 - 1.5
   //								 	110 - 2.5
   // Bit 3 (BT - Burst Type): 0 - Sequential, 1 - Interleaved
   // Bits 2-0 (BL - burst length):		001 - 2
   //									010 - 4
   //									011 - 8
   parameter MRS_SET = 15'b000000000100001; //BL=2,BT=S,LT=2 
   //

output [12:0] ddr_a;
output [3:0] ddr_cmd;
output [1:0] ddr_ba;
output [1:0] ddr_clk;
inout [15:0] ddr_dq;
output [1:0] ddr_dm;
inout [1:0] ddr_dqs;
output [31:0] ddr_data_out;
output ddr_cke;
output [63:0] ddr_test_out;
input [1:0] ddr_pll_in;
input ddr_pll_90;
input [3:0] ddr_cmd_in;
input [1:0] ddr_ba_in;
input [12:0] ddr_r_in;
input [8:0] ddr_c_in;
input [31:0] ddr_data_in;
input ddr_test_shift;
input ddr_test_clk;

reg [12:0] ddr_a;
reg [63:0] ddr_test_out;

// Pass PLL inputs directly to the memory
assign ddr_clk[1:0] = ddr_pll_in[1:0];

wire init_done;
wire [18:0] ddr_init_fsm_cmd;
wire [3:0] ddr_fsm_cmd;
wire [3:0] fsm_stat;
wire [3:0] fsm_next_stat;
wire [3:0] fsm_next_cmd_stat;

wire [1:0] ddr_dqs_rd;
wire fsm_op_done;
reg fsm_op_done_dly;
reg [3:0] fsm_cmd_i;
reg ddr_dqs_en;
reg ddr_cke;
reg [31:0] ddr_count;
reg [15:0] ddr_start_count; // Activated only for the first 200us
reg start_delay_done;
reg [1:0] ddr_dm_drv = 2'b00;	// Constant value as mask is not used
reg [3:0] vj_cmd_flag;
reg [3:0] ddr_dqs_count;
reg [15:0] ddr_dq_drv;
reg [2:0] ddr_read_lt;
reg [63:0] ddr_ram [0:31];	// Array of test registers
reg [15:0] ddr_data_out_buf[1:0];	// Buffer for output data
reg [7:0] ddr_test_count;
reg [4:0] ddr_index;
reg [14:0] ddr_mrs;			// This register stores MRS value



// DQS - output when read data, input with write data (for memory).
// Edge aligned with read data, centered in write data. Used to capture write data
assign ddr_dqs = (ddr_dqs_en)? {ddr_pll_in[0],ddr_pll_in[0]} : 2'bzz;

assign ddr_dq = (ddr_dqs_en)? ddr_dq_drv : 16'hzzzz; 

assign ddr_dm = ddr_dm_drv;		// Do not use mask

// Instantiate FSM's
ddr_init_fsm fsm_init_proc (ddr_mrs, ddr_clk[0], start_delay_done, ddr_count[7:0], ddr_init_fsm_cmd[18:0], init_done);
ddr_fsm fsm_proc (ddr_clk[0], init_done, fsm_cmd_i[3:0], ddr_fsm_cmd[3:0],fsm_next_cmd_stat[3:0], fsm_stat[3:0], fsm_next_stat[3:0], fsm_op_done);

// Commands logic - either from ddr_init_fsm during initialization
// or ddr_fsm during main process
assign {ddr_cmd[3:0], ddr_ba[1:0]} = (init_done)? {ddr_fsm_cmd[3:0],ddr_ba_in[1:0]} : ddr_init_fsm_cmd[18:13];

// To place read data in order depending on read latency
assign ddr_data_out[15:0] = ((ddr_mrs[6:4] == 3'b101) || (ddr_mrs[6:4] == 3'b110)) ? ddr_data_out_buf[0]:		// Latency 1.5 or 2.5
							((ddr_mrs[6:4] == 3'b010) || (ddr_mrs[6:4] == 3'b011)) ? ddr_data_out_buf[1]:	16'h0000;// Latency 2.0 or 3.0
							
assign ddr_data_out[31:16] = ((ddr_mrs[6:4] == 3'b101) || (ddr_mrs[6:4] == 3'b110))  ? ddr_data_out_buf[1]:	// Latency 1.5 or 2.5
							 ((ddr_mrs[6:4] == 3'b010) || (ddr_mrs[6:4] == 3'b011)) ? ddr_data_out_buf[0]: 16'h0000;	// Latency 2.0 or 3.0

// Block for data output during WRITE command
always @ (ddr_dqs_en or ddr_pll_90 or ddr_data_in[31:0])
  begin    
    ddr_dq_drv[15:0] = 16'hxxxx;
    if (ddr_dqs_en == 1'b1)
      case (ddr_pll_90)
        1'b1: ddr_dq_drv[15:0] = ddr_data_in[15:0];
        1'b0: ddr_dq_drv[15:0] = ddr_data_in[31:16];
      endcase
  end

// Block to select driver for DDR address lines. Either from ddr_init_fsm or
// ddr_fsm. If from ddr_fsm then either column or raw address
always @ (init_done or ddr_init_fsm_cmd[12:0] or ddr_r_in[12:0] or fsm_stat or ddr_c_in[8:0])    // Address lines multiplexor
  begin
    ddr_a[12:0] = 13'bxxxxxxxxxxxxx;
    if (init_done == 1'b0)
      ddr_a[12:0] = ddr_init_fsm_cmd[12:0];
    else
      case (fsm_stat)
        RA, MRS: ddr_a[12:0] = ddr_r_in[12:0];
        default: ddr_a[12:0] = {4'b0,ddr_c_in[8:0]};
      endcase  
  end

// Block for WRITE command to enable ddr_dqs_en  
always @ (fsm_op_done_dly or ddr_dqs_count or fsm_stat)
  begin
    ddr_dqs_en = 1'bx;
    if ((ddr_dqs_count > 4'h2) && (fsm_op_done_dly == 1'b1) && (fsm_stat == WRITE))
      ddr_dqs_en = 1'b1;
    else
      ddr_dqs_en = 1'b0;
  end
 
// Main logic contained in this always block 
always @ (posedge ddr_clk[0])
  begin
    fsm_op_done_dly <= fsm_op_done;
    
    if (ddr_start_count !== 16'h0010) // Start delay counter
      begin
        ddr_start_count <= ddr_start_count + 1'b1;
        start_delay_done <= 1'b0;
        ddr_mrs <= MRS_SET;
        ddr_cke <= 1'b0;
      end
    else
      begin
        start_delay_done <= 1'b1;
        ddr_cke <= 1'b1;
      end
      
    case (start_delay_done)		// Main counter
      1'b0:					// Do not count at this time
        ddr_count <= 32'h00000000;	
      1'b1:					// Count otherwise
        if ({fsm_stat, fsm_next_stat} != {AR,IDLE})
          ddr_count <= ddr_count + 1'b1;	// 30Ch is tREFI time
        else
          ddr_count <= ddr_count - 32'h0000030C;	// After Auto Refresh is completed need to zero counter
    endcase
     


    case (ddr_cmd_in)				// Set flags based on VJTAG commands
      WR_1: vj_cmd_flag <= WR_1;
      RD_1: vj_cmd_flag <= RD_1;
      MRS_SETUP: vj_cmd_flag <= MRS_SETUP;
    endcase  

      
    case ({vj_cmd_flag, fsm_stat, fsm_next_stat}) // Decode VJTAG commands
      {WR_1, IDLE, IDLE}: fsm_cmd_i <= ACT_T;	  // Write command
	  {WR_1, RA, RA}: fsm_cmd_i <= WRITE_T;
	  {WR_1, WRITE, WRITE}:
	    if (fsm_op_done)
		  fsm_cmd_i <= NOP_T;
		else
		  fsm_cmd_i <= PRE_T; 
      {WR_1, PC, PC}:
        begin
          fsm_cmd_i <= NOP_T;
          vj_cmd_flag <= NOP_VJ;		// Write cycle is done
         end
          		
	  {RD_1, IDLE, IDLE}: fsm_cmd_i <= ACT_T;	// Read command
      {RD_1, RA, RA}: fsm_cmd_i <= READ_T;
      {RD_1, READ, READ}:
        if (fsm_op_done)
          fsm_cmd_i <= NOP_T;
        else 
          fsm_cmd_i <= PRE_T;
      {RD_1, PC, PC}:
        begin
          fsm_cmd_i <= NOP_T;
          vj_cmd_flag <= NOP_VJ;		// Read cycle is done
         end
         
       {MRS_SETUP, IDLE, IDLE}: fsm_cmd_i <= MRS_T;	// MRS command
       {MRS_SETUP, MRS,IDLE}: 
         begin
           fsm_cmd_i <= NOP_T;
           vj_cmd_flag <= NOP_VJ;					// MRS setup done
           ddr_mrs[14:0] <= {ddr_ba_in[1:0],ddr_r_in[12:0]};		// To record updated MRS value
         end
              
      default:
        if (ddr_count > 32'h0000030C) 	// Need to do refresh
          case ({fsm_stat, fsm_next_stat})
            {IDLE,IDLE}: 	fsm_cmd_i <= REFA_T;
             default: 		fsm_cmd_i <= NOP_T;
          endcase   
	    else
	      fsm_cmd_i <= NOP_T;
    endcase
    
    if (ddr_cmd == READ_CMD)		// Logic for read latency counter
      ddr_read_lt <= READ_LT;
    else 
      if (ddr_read_lt != 3'b000)
        ddr_read_lt <= ddr_read_lt - 1'b1;   // Decrement read counter until it's zero  
        
  end

always @ (negedge ddr_pll_90)	// Edges of ddr_pll_90 signal used instead of ddr_dqs
  begin							// to read data
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b101,(READ_LT - 3'b010)})	// Latency 1.5
      ddr_data_out_buf[0] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b010,(READ_LT - 3'b011)})	// Latency 2.0
      ddr_data_out_buf[0] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b110,(READ_LT - 3'b011)})	// Latency 2.5
      ddr_data_out_buf[0] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b011,(READ_LT - 3'b100)}) 	// Latency 3.0
      ddr_data_out_buf[0] <= ddr_dq[15:0];
   
  end  

always @ (posedge ddr_pll_90)
  begin
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b101,(READ_LT - 3'b010)})	// Latency 1.5
      ddr_data_out_buf[1] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b010,(READ_LT - 3'b010)})	// Latency 2.0
      ddr_data_out_buf[1] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b110,(READ_LT - 3'b011)})	// Latency 2.5
      ddr_data_out_buf[1] <= ddr_dq[15:0];
    if ({ddr_mrs[6:4],ddr_read_lt} == {3'b011,(READ_LT - 3'b011)}) 	// Latency 3.0
      ddr_data_out_buf[1] <= ddr_dq[15:0];
      
  end



always @ (negedge ddr_clk[0])	// Used in WRITE commands
  begin
    if (fsm_op_done)
      ddr_dqs_count <= ddr_dqs_count + 1'b1;
    else
      ddr_dqs_count <= 4'h0;
  end
 
always @ (negedge ddr_test_shift)	// For test purposes - when data to be
  begin								// uploaded to vjtag_decoder block
    ddr_index <= ddr_index + 1'b1;
    ddr_test_out <= ddr_ram[ddr_index];
  end
  
always @ (posedge ddr_test_clk)		// Logic to record test data
  begin
    case ({vj_cmd_flag, fsm_stat, fsm_next_stat})
      {WR_1, IDLE, IDLE}: ddr_test_count <= 8'h00;
      {RD_1, IDLE, IDLE}: ddr_test_count <= 8'h00;
      {MRS_SETUP, IDLE, IDLE}: ddr_test_count <= 8'h00;
      default:
        if (ddr_test_count !== 8'h20)
          ddr_test_count <= ddr_test_count + 1'b1;
    endcase
      
    if ((ddr_test_count >= 8'h00) && (ddr_test_count < 8'h20))
      case (vj_cmd_flag)	// Data to be recorded can various depending on
        WR_1:				// requirements
          ddr_ram[ddr_test_count[4:0]] <= {ddr_test_count[7:0], fsm_stat[3:0],
          fsm_next_stat[3:0], fsm_cmd_i[3:0], ddr_a[12:0], 3'b000, 
          ddr_fsm_cmd[3:0], ddr_dqs_count[3:0], ddr_dqs_en, ddr_read_lt[2:0],  ddr_dq_drv[15:0]};
        RD_1:
          ddr_ram[ddr_test_count[4:0]] <= {2'b00, ddr_dqs[1:0], ddr_test_count[3:0], fsm_stat[3:0],
          fsm_next_stat[3:0], fsm_cmd_i[3:0], ddr_a[12:0], 3'b000, 
          ddr_fsm_cmd[3:0], 1'b0, ddr_read_lt[2:0], ddr_clk[0], 2'b00, ddr_pll_90, 1'b0, ddr_mrs[14:0]};
        default:
          ddr_ram[ddr_test_count[4:0]] <= {2'b00, ddr_dqs[1:0], ddr_test_count[3:0], fsm_stat[3:0],
          fsm_next_stat[3:0], fsm_cmd_i[3:0], ddr_a[12:0], 3'b000, 
          ddr_fsm_cmd[3:0], 1'b0, ddr_read_lt[2:0], ddr_clk[0], 2'b00, ddr_pll_90, 1'b0, ddr_mrs[14:0]};
  
      endcase
  end

endmodule
