// This file implements DDR SDRAM Main Operation Finite State Machine
// 


module ddr_fsm (
fsm_clk_in, fsm_start, fsm_cmd_in,
fsm_cmd_out, fsm_next_cmd_out, fsm_state, fsm_next_state, fsm_op_done
);
  
parameter [15:0]	// Timing requirements for FSM transitions in counter counts
	TRCDW	= 16'h0004,	//Row to Column delay min = 15nS (counter for WRITE operation)
	TRCDR = 16'h0004, //(16'h0002) Row to Column delay min = 15nS (counter for READ operation)
	TWR		= 16'h0002,	// Write recovery time min = 15nS
	TWR2	= 16'h0002,	// Time required to write 2 cycles burst
	TRP  = 16'h0002, // Row precharge time min = 15nS
	TRAS = 16'h0006,  // Row active time min = 40nS
	TRFC = 16'h0008,  // Auto Refresh to Active command period min = 70nS
	TRDPC = 16'h0001, // Delay between READ and PC commands
	TMRI = 16'h00C8;  // Wait for 200 cycles after MRS command
	
`include "parameters.v"

input fsm_clk_in;
input fsm_start;
input [3:0] fsm_cmd_in;
output [3:0] fsm_cmd_out;
output [3:0] fsm_next_cmd_out;
output [3:0] fsm_state;
output [3:0] fsm_next_state;
output fsm_op_done;

reg [3:0] fsm_cmd_out;
reg [15:0] fsm_state_counter = 16'h0000;

reg [3:0] fsm_next_cmd_out;
reg[3:0] fsm_state;
reg[3:0] fsm_next_state;
reg fsm_op_done;


always @ (fsm_cmd_in or fsm_state or fsm_state_counter)
  begin
	    fsm_next_state = IDLE;			// To make sure FSM is properly
	    fsm_next_cmd_out = NOP_CMD;		// initialized
	    fsm_op_done = 1'b0;
	    
	// State machine
	case (fsm_state)	// This case does not have default statement
  
    IDLE:
      if (fsm_state_counter == 16'h0000)     // Decode commands only after tRP time
        case (fsm_cmd_in)
          REFA_T:
            begin
		      fsm_next_state = AR;
		      fsm_next_cmd_out = REFA_CMD;
            end
          ACT_T:
            begin
              fsm_next_state = RA;
              fsm_next_cmd_out = ACT_CMD;
            end
          MRS_T:
            begin
              fsm_next_state = MRS;
              fsm_next_cmd_out = MRS_CMD;
            end
        endcase
        
	MRS: fsm_next_state = IDLE; 	// Only automatic route
	
	AR:
	  case (fsm_state_counter)
        TRFC: fsm_next_state = AR;
        16'h0000: fsm_next_state = IDLE;
        default: fsm_next_state = AR;
        endcase
	RA:
      case (fsm_cmd_in)
        READ_T:
          begin
            fsm_next_state = READ;
            fsm_op_done = 1'b1;
          end
        PRE_T: fsm_next_state = PC;
        WRITE_T:
          begin
            fsm_next_state = WRITE;
            fsm_op_done = 1'b1;	
          end
        NOP_T: fsm_next_state = RA;
        default: fsm_next_state = RA;
      endcase
	
	READ:
        case (fsm_cmd_in)
        READ_T:
          begin
            fsm_next_state = READ;
            fsm_op_done = 1'b1;
          end
        PRE_T: fsm_next_state = PC;
        NOP_T:
	      case (fsm_state_counter)   // tRCDR time
	        16'h0004:
	          begin
	            fsm_next_state = READ;
	            fsm_next_cmd_out = READ_CMD;   
	            fsm_op_done = 1'b1;
	          end
	        16'h0003: 
	          begin
	            fsm_next_state = READ;
	            fsm_op_done = 1'b1;
	          end
	        16'h0001: fsm_next_state = READ;
	        16'h0000: 
	          begin
	            fsm_next_state = RA; // If no commands sent after tRCD times out - 
	            fsm_next_cmd_out = ACT_CMD; // go back to Row active state
	          end 
	        default:
	          begin
	            fsm_next_state = READ;
	            fsm_op_done = 1'b1;
	          end 
	      endcase                      
        default:			// Automatic route is to return to RA state
          begin
            fsm_next_state = RA;
            fsm_next_cmd_out = ACT_CMD;
            fsm_op_done = 1'b1;
          end
      endcase
	WRITE:
      case (fsm_cmd_in)
        WRITE_T:
          begin
            fsm_next_state = WRITE;
            fsm_op_done = 1'b1;
          end
        PRE_T: fsm_next_state = PC;
        NOP_T:
          case (fsm_state_counter)   // tRCDW time
	        16'h0004:
	          begin
	            fsm_next_state = WRITE;
   	            fsm_next_cmd_out = WRITE_CMD;   // Insert two NOP cycles before 
   	            fsm_op_done = 1'b1;
   	          end
	        16'h0003:
	          begin
	            fsm_next_state = WRITE;
	            fsm_op_done = 1'b1;
	          end
	        16'h0001: fsm_next_state = WRITE;
	        16'h0000: 
	          begin
	            fsm_next_state = RA; // If no commands sent after tRCD times out - 
	            fsm_next_cmd_out = ACT_CMD; // go back to Row active state
	          end  
	        default:
	          begin
	            fsm_next_state = WRITE;
	            fsm_op_done = 1'b1;
	          end
	      endcase                
        default:
          begin
            fsm_next_state = RA;
            fsm_next_cmd_out = ACT_CMD;
            fsm_op_done = 1'b1;
          end
      endcase 	  
	PC:
	  case (fsm_state_counter)  
	    16'h0001:
	      begin
            fsm_next_state = PC;
	        fsm_next_cmd_out = PRE_CMD;// Need to wait for tWR,tRAS times
	      end
	    16'h0000: fsm_next_state = IDLE;
        default: fsm_next_state = PC;
      endcase
  endcase
end
  

always @(negedge fsm_clk_in)		// Build FSM flip-flops
  begin
    if (!fsm_start)
      begin
        fsm_state <= IDLE;
        fsm_cmd_out <= NOP_CMD;
        fsm_state_counter <= 16'h0000;
      end
    else
      begin
        fsm_state <= fsm_next_state;
        fsm_cmd_out <= fsm_next_cmd_out;
      end
      
        case ({fsm_state,fsm_next_state})
          {RA,WRITE}: fsm_state_counter <= TRCDW;
          {RA,READ}:  fsm_state_counter <= TRCDR;
          {WRITE,PC}: fsm_state_counter <= TWR;
          {RA,PC}:    fsm_state_counter <= TRAS;
          {READ,PC}:  fsm_state_counter <= TRDPC;
          {PC,IDLE}:  fsm_state_counter <= TRP; 
          {IDLE,AR}:  fsm_state_counter <= TRFC;
          {IDLE,MRS}: fsm_state_counter <= TMRI;	
          default:
            if (fsm_state_counter !== 16'h0000)
              fsm_state_counter <= fsm_state_counter - 1'b1;
        endcase  
  end 
 
endmodule



